import algos.*
import experiments.GCCLazy
import org.jgrapht.GraphTests
import org.jgrapht.graph.SimpleGraph
import java.io.File
import java.io.FileOutputStream
import java.io.PrintStream

/**
 * argument format:
 *
 *    [0] output path: string
 *    [1] graph path: string
 *    [2] k: int
 *    [3] type of algo: string
 *    [4] problem ID: string
 *    [5] timeout in ms: int
 *    [6] initial solution: string
 *    [7] additional time in ms: int
 */
fun main(args: Array<String>) {
    // get Input Parameters
    val pathOutput = args[0]
    val pathGraph = args[1]
    val k = args[2].toInt()
    val algoType = args[3]
    val problemID = args[4]
    val timeoutMS = args[5].toLong()
    // initial solution (provided by greedy++)
    val initSol = args[6]
    val addTime = args[7].toLong()
    // create PrintStream for Output File
    val out = PrintStream(FileOutputStream(File(pathOutput), true))
    // create Graph from File
    val graph = importSimpleGraphFromFile(pathGraph)
    // get biggest connected component
    if (!GraphTests.isConnected(graph)) {
        reduceToBiggestConnectedSubGraph(graph)
    }
    // Create the correct Algorithm
    val algo = chooseAlgo(algoType, graph, getVertexSet(initSol))
    if (algo != null) {
        // run the algorithm for the given k with the given timeout
        println("Working on problem $problemID...")
        var completed = false
        val memory = measureMemoryUsageOfTask {
            completed = runTaskWithTimeout(timeoutMS) {
                algo.solve(k)
            }
        }
        if (completed) {
            println("Problem $problemID Completed")
        } else {
            println("Problem $problemID Timeout")
        }
        printInfoCSV(problemID, out, algo, memory, !completed, algoType, pathGraph, addTime)
        // val dens = 2*algo.edgeCount.toDouble()/(algo.vertCount.toDouble()*(algo.vertCount-1).toDouble())
        // println("number of vertices ${algo.vertCount} number of edges ${algo.edgeCount} diameter ${algo.diameter} density $dens")
    } else {
        println("Fourth argument needs to be one of ${listOf("Basic", "CentImp", "GCC", "ILP")}")
    }
    out.close()
}

fun <V, E> chooseAlgo(algoType: String, graph: SimpleGraph<V, E>, initSol: Set<V>): Algo<V, E>? {
    return when (algoType) {
        "Plain" -> Plain(graph)
        "CI" -> GCC(graph, useUpperBound = false, useDomVerts = false, useGivenSol = false)
        "UB" -> GCC(graph, useUpperBound = true, useDomVerts = false, useGivenSol = false)
        "DV" -> GCC(graph, useUpperBound = true, useDomVerts = true, useGivenSol = false)
        "DVISStart" -> GCC(graph, useUpperBound = false, useDomVerts = true, useGivenSol = true, initSol)
        "DVind" -> GCCiter(graph, useUpperBound = true, useDomVerts = true, useGivenSol = false, glob = false)
        "DVglob" -> GCCiter(graph, useUpperBound = true, useDomVerts = true, useGivenSol = false, glob = true)
        "DVindSp" -> GCCiterSparse(graph, useUpperBound = false, useDomVerts = true, useGivenSol = false, glob = false)
        "DVglobSp" -> GCCiterSparse(graph, useUpperBound = false, useDomVerts = true, useGivenSol = false, glob = true)
        "DVLazy" -> GCCLazy(graph, useUpperBound = true, useDomVerts = true, useGivenSol = false)
        "ILPold" -> ILPold(graph)
        "ILPnew" -> ILPnew(graph)
        "ILPind" -> ILPIterNoDists(graph, glob = false)
        "ILPglob" -> ILPIterNoDists(graph, glob = true)
        else -> null
    }
}

fun <V, E> printInfoPretty(out: PrintStream, algo: Algo<V, E>, printAdditionalInfo: Boolean = true) {
    out.println("Results with Optimized Algorithm for size ${algo.k}:")
    out.println("Time for Prep: ${prettyTime(algo.prepTime)}")
    out.println("Time for Algo: ${prettyTime(algo.algoTime)}")
    out.println("Result: (${algo.lastResult.second}, ${algo.lastResult.third}) ${algo.lastResult.first}")
    if (printAdditionalInfo) {
        out.println("\nAdditional Info:")
        out.println("Graph Info: #Vertices = ${algo.vertCount} #Edges = ${algo.edgeCount}")
        out.println("Graph Diameter: ${algo.diameter}")
        if (algo is GCC) {
            out.println("Upper Bound: (${algo.lastUpperBound.second}, ${algo.lastUpperBound.third}) ${algo.lastUpperBound.first}")
            out.println("Sufficient Vertices: (${algo.sufficientVertices.size}) ${algo.sufficientVertices}")
            out.println("Checked Subsets: ${algo.checkedSubsets}")
        }
        if (algo is Plain) {
            out.println("Checked Subsets: ${algo.checkedSubsets}")
        }
    }
}

/**
 * Columns of the output file:
 *
 *    [00] problem ID: string
 *    [01] # of vertices: int
 *    [02] # of edges: int
 *    [03] diameter: int
 *    [04] k: int
 *    [05] maximum used memory in MiB: int
 *    [06] time for preparations in ms (also includes added time): int
 *    [07] time for algorithm in ms: int
 *    [08] total time: int
 *    [09] # of sufficient vertices: int
 *    [10] centrality of upper bound: int
 *    [11] centrality of result: int
 *    [12] normalized centrality of result: float
 *    [13] list of vertices from the result: string
 *    [14] name of the algorithm: string
 *    [15] the path to the graph that was used: string
 *    [16] number of iterations used in specific ILPs: int
 */
fun <V, E> printInfoCSV(problemID: String, out: PrintStream, algo: Algo<V, E>, memory: Long, timedOut: Boolean, algString: String, pathGraph: String, addTime: Long) {
    if (!timedOut) {
        out.println("$problemID;${algo.vertCount};${algo.edgeCount};${algo.diameter};${algo.k};$memory;${algo.prepTime + addTime};${algo.algoTime};${algo.totalTime + addTime};${algo.sufficientVertices.size};${algo.lastUpperBound.second};${algo.lastResult.second};${algo.lastResult.third};${algo.lastResult.first};$algString;$pathGraph;${algo.iterations}")
    } else {
        out.println("$problemID;${algo.vertCount};${algo.edgeCount};null;${algo.k};null;null;null;null;null;null;null;null;null;$algString;$pathGraph;null")
    }
}
