package algos

import calcCombinedCentrality
import combineDistances
import org.jgrapht.Graphs
import org.jgrapht.graph.SimpleGraph
import java.util.*
import kotlin.math.max

abstract class Algo<V, E>(protected val graph: SimpleGraph<V, E>) {

    // Variables for storing information about the algorithm
    val vertCount = graph.vertexSet().size
    val edgeCount = graph.edgeSet().size

    /**
     * The diameter of the given graph
     */
    var diameter: Int = 0
        protected set

    /**
     * The value for k used in the last solve operation
     */
    var k = 0
        protected set

    /**
     * The last result calculated by the algorithm
     */
    var lastResult = Triple<Set<V>, Int, Float>(HashSet(), 0, 0f)
        protected set

    /**
     * Time in ms for preparations
     */
    var prepTime = 0L
        private set

    /**
     * Time in ms for main part of the algorithm
     */
    var algoTime = 0L
        private set

    /**
     * Total time in ms of last solve operation
     */
    val totalTime: Long
        get() = prepTime + algoTime

    var lastUpperBound = Triple<Set<V>, Int, Float>(HashSet(), 0, 0f)
        protected set

    var checkedSubsets = 0
        protected set

    var sufficientVertices = HashSet<V>(vertCount)
        protected set

    /**
     * number of iterations in specific ILPs
     */
    var iterations: Int = 0
        protected set

    var bfsExpansions: Int = 0

    /**
     * amount of time each iteration took
     */
    var iterTime: MutableList<Long> = mutableListOf()

    /**
     * number of times a centrality improvement of any vertex was calculated
     */
    var centImpCalcs = 0
        protected set

    private var tempTime = 0L

    protected var vertexToOrder: MutableMap<V, Int> = HashMap(0)
    protected var vertexList: List<V> = ArrayList(0)
    protected var distanceMatrix: Array<IntArray> = Array(0) { IntArray(0) }
    protected var neighborhoods: Array<Set<Int>> = Array(0) { setOf() }

    // length of a longest shortest path starting in this vertex  (initialized with 0)
    protected var eccentricity: IntArray = IntArray(vertCount) { 0 }

    enum class NextAction { UP, STAY, DOWN }

    fun solve(k: Int): Triple<Set<V>, Int, Float> {
        if (k < 1 || k >= vertCount) {
            return Triple(HashSet(), 0, 0f)
        }

        this.k = k
        // preparations
        tempTime = System.currentTimeMillis()
        preparations()
        prepTime = System.currentTimeMillis() - tempTime
        // calculate result
        tempTime = System.currentTimeMillis()
        lastResult = calcResult()
        algoTime = System.currentTimeMillis() - tempTime

        return lastResult
    }

    protected fun createVertexLists() {
        vertexList = graph.vertexSet().toList()
        vertexToOrder = HashMap(k)
        for (i in vertexList.indices) {
            vertexToOrder[vertexList[i]] = i
        }
    }

    protected fun calcNeighborhoods() {
        neighborhoods = Array(vertCount) { vi ->
            val res = HashSet<Int>()
            Graphs.neighborSetOf(graph, vertexList[vi]).forEach { v ->
                res.add(vertexToOrder[v]!!)
            }
            res
        }
    }

    protected fun createDistanceMatrix() {
        distanceMatrix = Array(vertCount) { vi ->
            val res = IntArray(vertCount) { -1 }
            val q: Queue<Int> = LinkedList()
            q.offer(vi)
            res[vi] = 0
            var ecc = 0
            while (q.isNotEmpty()) {
                val cur = q.poll()
                for (ni in neighborhoods[cur]) {
                    if (res[ni] == -1) {
                        res[ni] = res[cur] + 1
                        ecc = max(ecc, res[ni])
                        q.offer(ni)
                    }
                }
            }
            eccentricity[vi] = ecc
            diameter = max(diameter, ecc)
            res
        }
    }

    protected fun calcUpperBoundGreedy(): Pair<Int, IntArray> {
        val indices = IntArray(k) { -1 }
        var i = 1
        indices[0] = vertexList.indices.minByOrNull { distanceMatrix[it].sum() }!!
        val distances = Array(k) { IntArray(vertCount) }
        distances[0] = distanceMatrix[indices[0]]
        var cent = distances[0].sum()
        while (i < k) {
            for (vi in vertexList.indices) {
                if (vi in indices) continue
                val temp = calcCombinedCentrality(distances[i - 1], distanceMatrix[vi])
                if (temp < cent) {
                    indices[i] = vi
                    cent = temp
                    combineDistances(distances[i - 1], distanceMatrix[vi], distances[i])
                }
            }
            i++
        }
        return Pair(cent, indices)
    }

    protected abstract fun preparations()

    protected abstract fun calcResult(): Triple<Set<V>, Int, Float>
}
