package algos

import checkIfSubset
import gnu.trove.list.array.TIntArrayList
import gnu.trove.set.hash.TIntHashSet
import org.jgrapht.graph.SimpleGraph
import java.util.*
import kotlin.math.max

class GCCiterSparse<V, E>(
    graph: SimpleGraph<V, E>,
    val useUpperBound: Boolean,
    val useDomVerts: Boolean,
    val useGivenSol: Boolean,
    val sol: Set<V> = setOf(),
    val glob: Boolean
) : Algo<V, E>(graph) {

    private var initialSufficientIndices: MutableSet<Int> = HashSet(vertCount)

    override fun preparations() {
        createVertexLists()
        calcNeighborhoods()
        calcDiameterUpperBound()
        if (useDomVerts) {
            calcSufficientVertexSet()
        }
    }

    private fun calcDiameterUpperBound() {
        val queue = LinkedList(listOf(0))
        val visited = HashSet(setOf(0))
        var dist = -1
        while (queue.isNotEmpty()) {
            dist++
            for (i in 0 until queue.size) {
                val vi = queue.pollFirst()
                for (ni in neighborhoods[vi]) {
                    if (ni !in visited) {
                        queue.addLast(ni)
                        visited.add(ni)
                    }
                }
            }
        }
        diameter = (dist * 2) + 1
    }

    private fun calcSufficientVertexSet() {
        val result = initialSufficientIndices
        val eliminated = HashSet<Int>(vertCount)
        for (vi in vertexList.indices) {
            if (vi in eliminated) continue
            result.add(vi)
            // find all vertices that are dominated by vi
            val vertSet = HashSet<Int>(vertCount)
            // start by finding vertices that fulfill the first property
            for (neigh in neighborhoods[vi]) {
                vertSet.addAll(neighborhoods[neigh])
                if (neigh in eliminated) continue
                // check first property
                if (checkIfSubset(neighborhoods[neigh], vi, neighborhoods[vi])) {
                    result.remove(neigh)
                    eliminated.add(neigh)
                }
            }
            vertSet.removeAll(neighborhoods[vi])
            vertSet.remove(vi)
            // now find vertices that fulfill the second property
            for (vin in vertSet) {
                if (vin in eliminated) continue
                // check second property
                if (neighborhoods[vin].size == neighborhoods[vi].size && neighborhoods[vin] == neighborhoods[vi]) {
                    result.remove(vin)
                    eliminated.add(vin)
                }
            }
        }
        sufficientVertices = HashSet(initialSufficientIndices.size)
        for (vi in initialSufficientIndices) {
            sufficientVertices.add(vertexList[vi])
        }
    }

    override fun calcResult(): Triple<Set<V>, Int, Float> {
        // immediately go to checkEverySubset() if useDominatingVertices is false
        return if (!useDomVerts || k < initialSufficientIndices.size) {
            doIterations()
        } else {
            // if the number of sufficient vertices is not larger than k the solution contains
            // all sufficient vertices plus any k - (# of sufficient vertices) additional vertices
            val result = HashSet<V>(k)
            for (vi in initialSufficientIndices) {
                result.add(vertexList[vi])
            }
            for (vert in vertexList) {
                if (result.size >= k) break
                if (vert !in result) {
                    result.add(vert)
                }
            }
            return Triple(result, vertCount - k, 1f)
        }
    }

    // initialize sparseMatrix with distances of 0, 1 and 2
    private val sparseMatrix = Array(vertCount) { Array(0) { TIntArrayList(0) } }
    private val dValues = IntArray(vertCount) { 2 }

    private fun sparseSum(vi: Int): Int {
        val cur = sparseMatrix[vi]
        var amount = vertCount
        var sum = 0
        for (d in 0 until diameter) {
            sum += cur[d].size() * d
            amount -= cur[d].size()
        }
        sum += dValues[vi] * amount
        return sum
    }

    private fun combineSparseDistances(fromArray: IntArray, fromVertex: Int, to: IntArray): Int {
        var sum = 0
        for (vi in 0 until vertCount) {
            to[vi] = fromArray[vi]
            sum += to[vi]
        }
        for (d in 0 until diameter) {
            for (vi in sparseMatrix[fromVertex][d]) {
                if (d < to[vi]) {
                    sum -= to[vi] - d
                    to[vi] = d
                }
            }
        }
        return sum
    }

    private fun calcCombinedSparseCentrality(fromArraySum: Int, fromArray: IntArray, fromVertex: Int): Int {
        var sum = fromArraySum
        for (d in 0 until diameter) {
            for (vi in sparseMatrix[fromVertex][d]) {
                if (d < fromArray[vi]) {
                    sum -= fromArray[vi] - d
                }
            }
        }
        return sum
    }

    private fun doIterations(): Triple<Set<V>, Int, Float> {
        // initialize sparseMatrix with distances of 0, 1 and 2
        for (vi in 0 until vertCount) {
            sparseMatrix[vi] = Array(diameter) { dist ->
                when (dist) {
                    0 -> {
                        val res = TIntArrayList(1)
                        res.add(vi)
                        res
                    }
                    1 -> {
                        val res = TIntArrayList(neighborhoods[vi].size)
                        neighborhoods[vi].forEach { res.add(it) }
                        res
                    }
                    else -> TIntArrayList(1)
                }
            }
        }
        // create breadth first search iterators
        val bfs = Array(vertCount) {
            object : Iterator<Any> {
                var queue = sparseMatrix[it][1]

                override fun hasNext(): Boolean = !queue.isEmpty

                override fun next(): Any {
                    val dist = dValues[it]
                    val newQueue = TIntHashSet(queue.size())
                    for (q in queue) {
                        for (nb in neighborhoods[q]) {
                            newQueue.add(nb)
                        }
                    }
                    for (vi in sparseMatrix[it][dist - 2]) {
                        newQueue.remove(vi)
                    }
                    for (vi in sparseMatrix[it][dist - 1]) {
                        newQueue.remove(vi)
                    }
                    val distArray = TIntArrayList(newQueue.size())
                    for (ui in newQueue) {
                        distArray.add(ui)
                        if (dValues[ui] <= dist) {
                            sparseMatrix[ui][dist].add(it)
                        }
                    }
                    newQueue.clear()
                    sparseMatrix[it][dist] = distArray
                    dValues[it]++
                    queue = distArray
                    return 0
                }
            }
        }
        val increaseD = HashSet<Int>()
        var solved = false
        var lastSolution: Pair<Set<Int>, Int>
        do {
            iterations++
            val curTime = System.currentTimeMillis()
            lastSolution = checkEverySubSet()
            iterTime.add(System.currentTimeMillis() - curTime)
            // check which vertices have an undefined distance to the solution and are not done with bfs
            val solDists = lastSolution.first.associateWith { vi ->
                IntArray(vertCount) { max(dValues[vi], dValues[it]) }
            }
            for (v1 in lastSolution.first) {
                for (d in 0 until diameter) {
                    for (v2 in sparseMatrix[v1][d]) {
                        solDists[v1]!![v2] = d
                    }
                }
            }
            for (v1 in 0 until vertCount) {
                if (!bfs[v1].hasNext()) continue
                val v2 = lastSolution.first.minByOrNull { solDists[it]!![v1] }!!
                if (solDists[v2]!![v1] >= max(dValues[v1], dValues[v2])) {
                    if (glob) {
                        for (ui in 0 until vertCount) {
                            if (bfs[ui].hasNext()) {
                                increaseD.add(ui)
                            }
                        }
                        break
                    } else {
                        increaseD.add(v1)
                    }
                }
            }
            if (increaseD.isEmpty()) {
                solved = true
            } else {
                // increase d value for the necessary vertices and update the sparse matrix
                for (vi in increaseD) {
                    bfs[vi].next()
                    bfsExpansions++
                }
                increaseD.clear()
            }
        } while (!solved)

        val normCentrality = (vertCount - k).toFloat() / lastSolution.second.toFloat()
        val bestSet = HashSet<V>(lastSolution.first.size)
        lastSolution.first.forEach { bestSet.add(vertexList[it]) }
        return Triple(bestSet, lastSolution.second, normCentrality)
    }

    private fun checkEverySubSet(): Pair<Set<Int>, Int> {
        // ---------------------- RESET VARIABLES ----------------------
        checkedSubsets = 0

        // ---------------------- CALCULATE INITIAL SOLUTION ----------------------
        var bestIndices = IntArray(k)
        var bestCentrality = 0
        if (useUpperBound) {
            // use greedy upper bound
            val ub = calcUpperBoundGreedy()
            bestCentrality = ub.first
            bestIndices = ub.second
        } else if (useGivenSol && sol.size == k) {
            // use given solution
            var ind = 0
            for (vertex in sol) {
                bestIndices[ind] = vertexToOrder[vertex]!!
                ind += 1
            }
            val bestDists = IntArray(vertCount) { i ->
                bestIndices.minOfOrNull { vertI ->
                    distanceMatrix[vertI][i]
                }!!
            }
            bestCentrality = bestDists.sum()
        } else {
            // set initial solution to something that is bigger than theoretical maximum
            bestCentrality = vertCount * vertCount
        }
        lastUpperBound = Triple(
            bestIndices.map { vertexList[it] }.toSet(),
            bestCentrality,
            (vertCount - k).toFloat() / bestCentrality.toFloat()
        )

        // ---------------------- INITIALIZE VARIABLES FOR ITERATING OVER THE SUBSETS ----------------------
        var nextAction = NextAction.STAY
        val setIndices = IntArray(k) { -1 }
        var nextI = 0
        var spacesLeft = k - nextI
        val centImprovements = Array(k) { IntArray(vertCount) }
        val vertOrder = Array(k) { LinkedList<Int>() }

        // ---------------------- INITIALIZE VARIABLES FOR DYNAMICALLY SAVING IMPORTANT VALUES ----------------------
        var curVI: Int
        var curCentrality = 0
        val centralities = IntArray(k)
        val tempDistances = Array(k) { IntArray(vertCount) }
        val centImprovementSums = IntArray(k)
        val topImprovementVerts: Array<Deque<Int>> = Array(k) { LinkedList() }

        // ---------------------- INITIAL ORDERING ----------------------
        // choose which vertices to start with
        val initialVertexIndices = if (useDomVerts) {
            initialSufficientIndices
        } else {
            vertexList.indices
        }
        // calculate the initial centrality improvements for every vertex
        for (vi in initialVertexIndices) {
            centImprovements[nextI][vi] = (vertCount * vertCount) - sparseSum(vi)
            vertOrder[nextI].add(vi)
        }
        vertOrder[nextI].sortByDescending { vi -> centImprovements[nextI][vi] }
        // put the first k vertices from the ordering into topImprovementVerts
        for (i in 0 until spacesLeft) {
            topImprovementVerts[nextI].addLast(vertOrder[nextI].pollFirst())
        }

        // ---------------------- START OF SEARCH TREE ----------------------

        if (k == 1) {
            // handle special case of size == 1
            for (vi in topImprovementVerts[0]) {
                vertOrder[0].add(vi)
            }
            for (vi in vertOrder[0]) {
                curCentrality = sparseSum(vi)
                checkedSubsets++
                if (curCentrality < bestCentrality) {
                    bestCentrality = curCentrality
                    bestIndices = intArrayOf(vi)
                }
            }
        } else {
            while (nextI > 0 || nextAction != NextAction.UP) {
                when (nextAction) {
                    NextAction.UP -> {
                        // clear variables
                        vertOrder[nextI].clear()
                        centImprovementSums[nextI] = 0
                        topImprovementVerts[nextI].clear()
                        // go back to previous Index
                        nextI--
                        spacesLeft = k - nextI
                        curCentrality = centralities[nextI]
                        // check if there are enough vertices left to keep going
                        if (topImprovementVerts[nextI].size == spacesLeft) {
                            nextAction = NextAction.STAY
                        }
                        continue
                    }
                    NextAction.STAY -> {
                        if (nextI > 0) {
                            // CENTRALITY LOWER BOUND
                            // check if the sum is too low
                            if (centralities[nextI - 1] - centImprovementSums[nextI] >= bestCentrality) {
                                nextAction = NextAction.UP
                                continue
                            } else {
                                // MINIMAL CENTRALITY THRESHOLD
                                // remove vertices with an improvement that is too low
                                var mc = centImprovementSums[nextI] - centImprovements[nextI][topImprovementVerts[nextI].last()]
                                if (curCentrality - mc > bestCentrality) {
                                    mc = curCentrality - mc - bestCentrality
                                    while (vertOrder[nextI].isNotEmpty() && centImprovements[nextI][vertOrder[nextI].last()] <= mc) {
                                        vertOrder[nextI].pollLast()
                                    }
                                    if (vertOrder[nextI].isEmpty() && centImprovements[nextI][topImprovementVerts[nextI].last()] <= mc) {
                                        nextAction = NextAction.UP
                                        continue
                                    }
                                }
                            }
                        }
                        // fill empty spot with next vertex
                        curVI = topImprovementVerts[nextI].pollFirst()!!
                        setIndices[nextI] = curVI
                        // calculate curCentrality and update tempDistances
                        curCentrality = if (nextI == 0) {
                            combineSparseDistances(dValues, curVI, tempDistances[nextI])
                        } else {
                            combineSparseDistances(tempDistances[nextI - 1], curVI, tempDistances[nextI])
                        }
                        centralities[nextI] = curCentrality
                        // update topImprovements and centImprovementSums
                        centImprovementSums[nextI] -= centImprovements[nextI][curVI]
                        if (vertOrder[nextI].isNotEmpty()) {
                            val top = vertOrder[nextI].pollFirst()
                            topImprovementVerts[nextI].addLast(top)
                            centImprovementSums[nextI] += centImprovements[nextI][top]
                        }
                        nextAction = NextAction.DOWN
                        continue
                    }
                    NextAction.DOWN -> {
                        // go to next Index
                        nextI++
                        spacesLeft = k - nextI
                        if (spacesLeft > 1) {
                            // PARENT THRESHOLD
                            // only add vertices that had a good enough centrality in the previous iteration
                            val last = topImprovementVerts[nextI - 1].pollLast()!!
                            var pt = centImprovementSums[nextI - 1] - centImprovements[nextI - 1][last] - centImprovements[nextI - 1][topImprovementVerts[nextI - 1].last()]
                            topImprovementVerts[nextI - 1].addLast(last)
                            pt = if (nextI > 1 && curCentrality - pt > bestCentrality) {
                                curCentrality - pt - bestCentrality
                            } else 0
                            // calculate centImprovements and add to vertOrder
                            for (vi in topImprovementVerts[nextI - 1]) {
                                if (centImprovements[nextI - 1][vi] <= pt) break
                                centImprovements[nextI][vi] = curCentrality - calcCombinedSparseCentrality(
                                    curCentrality,
                                    tempDistances[nextI - 1],
                                    vi
                                )
                                centImpCalcs++
                                vertOrder[nextI].add(vi)
                            }
                            for (vi in vertOrder[nextI - 1]) {
                                if (centImprovements[nextI - 1][vi] <= pt) break
                                centImprovements[nextI][vi] = curCentrality - calcCombinedSparseCentrality(
                                    curCentrality,
                                    tempDistances[nextI - 1],
                                    vi
                                )
                                centImpCalcs++
                                vertOrder[nextI].add(vi)
                            }

                            if (vertOrder[nextI].size < spacesLeft) {
                                nextAction = NextAction.UP
                            } else {
                                // sort vertOrder[nextI]
                                vertOrder[nextI].sortByDescending { vi -> centImprovements[nextI][vi] }
                                // fill topImprovements
                                for (i in 0 until spacesLeft) {
                                    val top = vertOrder[nextI].pollFirst()
                                    centImprovementSums[nextI] += centImprovements[nextI][top]
                                    topImprovementVerts[nextI].addLast(top)
                                }
                                nextAction = NextAction.STAY
                            }
                        } else {
                            // handle special case with spacesLeft == 1
                            val checkVertex = { vi: Int ->
                                setIndices[nextI] = vi
                                // PARENT THRESHOLD
                                // check if improvement from previous step is big enough
                                if (k <= 2 || curCentrality - centImprovements[nextI - 1][vi] < bestCentrality) {
                                    // calc and check centrality
                                    val cent = calcCombinedSparseCentrality(curCentrality, tempDistances[nextI - 1], vi)
                                    if (cent < bestCentrality) {
                                        // update best centrality
                                        bestCentrality = cent
                                        bestIndices = setIndices.clone()
                                    }
                                    checkedSubsets++
                                    true
                                } else {
                                    // stop checking here because the improvements after this will all be smaller
                                    false
                                }
                            }
                            for (vi in topImprovementVerts[nextI - 1]) {
                                if (!checkVertex(vi)) break
                            }
                            for (vi in vertOrder[nextI - 1]) {
                                if (!checkVertex(vi)) break
                            }
                            nextAction = NextAction.UP
                        }
                        continue
                    }
                }
            }
        }
        return Pair(bestIndices.toSet(), bestCentrality)
    }
}
