package algos

import checkIfSubset
import gurobi.GRB
import gurobi.GRBEnv
import gurobi.GRBLinExpr
import gurobi.GRBModel
import org.jgrapht.graph.SimpleGraph
import java.util.*

class ILPIterNoDists<V, E>(graph: SimpleGraph<V, E>, val glob: Boolean) : Algo<V, E>(graph) {

    private var initialSufficientIndices: MutableSet<Int> = HashSet(vertCount)

    override fun preparations() {
        createVertexLists()
        calcNeighborhoods()
        calcSufficientVertexSet()
    }

    private fun calcSufficientVertexSet() {
        val result = initialSufficientIndices
        val eliminated = HashSet<Int>(vertCount)
        for (vi in vertexList.indices) {
            if (vi in eliminated) continue
            result.add(vi)
            // find all vertices that are dominated by vi
            val vertSet = HashSet<Int>(vertCount)
            // start by finding vertices that fulfill the first property
            for (neigh in neighborhoods[vi]) {
                vertSet.addAll(neighborhoods[neigh])
                if (neigh in eliminated) continue
                // check first property
                if (checkIfSubset(neighborhoods[neigh], vi, neighborhoods[vi])) {
                    result.remove(neigh)
                    eliminated.add(neigh)
                }
            }
            vertSet.removeAll(neighborhoods[vi])
            vertSet.remove(vi)
            // now find vertices that fulfill the second property
            for (vin in vertSet) {
                if (vin in eliminated) continue
                // check second property
                if (neighborhoods[vin].size == neighborhoods[vi].size && neighborhoods[vin] == neighborhoods[vi]) {
                    result.remove(vin)
                    eliminated.add(vin)
                }
            }
        }
        sufficientVertices = HashSet(initialSufficientIndices.size)
        for (vi in initialSufficientIndices) {
            sufficientVertices.add(vertexList[vi])
        }
    }

    override fun calcResult(): Triple<Set<V>, Int, Float> {
        return if (k < initialSufficientIndices.size) {
            calcResultILP()
        } else {
            val result = HashSet<V>(k)
            for (vi in initialSufficientIndices) {
                result.add(vertexList[vi])
            }
            for (vert in vertexList) {
                if (result.size >= k) break
                if (vert !in result) {
                    result.add(vert)
                }
            }
            return Triple(result, vertCount - k, 1f)
        }
    }

    fun calcResultILP(): Triple<Set<V>, Int, Float> {
        // setup variables for dynamic distance calculations
        val prevVerts = Array(vertCount) {
            val res = HashSet(neighborhoods[it])
            res.add(it)
            res
        }
        val vertQueues = Array(vertCount) { LinkedList(neighborhoods[it]) }
        val finishedBFS = Array(vertCount) { false }
        val dValues = IntArray(vertCount) { 2 }

        // setup environment
        val env = GRBEnv()
        env.set(GRB.IntParam.OutputFlag, 0)
        env.set(GRB.IntParam.Threads, 1)
        env.start()
        // create model
        val model = GRBModel(env)

        // create initial variables and constraints

        // initial variables (d(v)=2 for each vertex)
        // if a vertex is not sufficient, we do not create the variable x_{v,0} since this vertex is not in the solution
        for (i in 0 until vertCount) {
            if (i in initialSufficientIndices) {
                model.addVar(0.0, 1.0, 0.0, GRB.BINARY, "x_${i}_0")
            }
            model.addVar(0.0, 1.0, 0.0, GRB.BINARY, "x_${i}_1")
            model.addVar(0.0, 1.0, 0.0, GRB.BINARY, "x_${i}_2")
        }

        model.update()

        // initial constraints (d(v)=2 for each vertex)
        val term0 = GRBLinExpr()
        for (i in 0 until vertCount) {
            val term = GRBLinExpr()
            if (i in initialSufficientIndices) {
                term0.addTerm(1.0, model.getVarByName("x_${i}_0"))
                term.addTerm(1.0, model.getVarByName("x_${i}_0"))
            }
            term.addTerm(1.0, model.getVarByName("x_${i}_1"))
            term.addTerm(1.0, model.getVarByName("x_${i}_2"))
            model.addConstr(term, GRB.EQUAL, 1.0, "sum vertex$i")
        }
        model.addConstr(term0, GRB.EQUAL, k.toDouble(), "choose k vertices")

        // distance constraints

        for (i in 0 until vertCount) {
            val term = GRBLinExpr()
            // only iterate over neighbors in the beginning
            for (j in neighborhoods[i]) {
                if (j in initialSufficientIndices) {
                    term.addTerm(1.0, model.getVarByName("x_${j}_0"))
                }
            }
            model.addConstr(model.getVarByName("x_${i}_1"), GRB.LESS_EQUAL, term, "ecc $i 1")
        }

        // objective function d(v)=2 for each vertex)
        val termi = GRBLinExpr()
        for (i in 0 until vertCount) {
            if (i in initialSufficientIndices) {
                for (j in 0 until 3) {
                    termi.addTerm(j.toDouble(), model.getVarByName("x_${i}_$j"))
                }
            } else {
                for (j in 1 until 3) {
                    termi.addTerm(j.toDouble(), model.getVarByName("x_${i}_$j"))
                }
            }
        }
        model.setObjective(termi, GRB.MINIMIZE)

        val increaseD = HashSet<Int>()
        var solved = false
        do {
            // solve
            iterations++
            val curTime = System.currentTimeMillis()
            model.optimize()
            iterTime.add(System.currentTimeMillis() - curTime)
            // check if we have a solution
            for (vi in 0 until vertCount) {
                val viVar = model.getVarByName("x_${vi}_${dValues[vi]}")
                if (!finishedBFS[vi] && viVar.get(GRB.DoubleAttr.X) > 0.99) {
                    if (glob) {
                        for (ui in 0 until vertCount) {
                            if (!finishedBFS[ui]) {
                                increaseD.add(ui)
                                dValues[ui]++
                            }
                        }
                        break
                    } else {
                        increaseD.add(vi)
                        dValues[vi]++
                    }
                }
            }

            if (increaseD.isEmpty()) {
                // solution was found
                solved = true
            } else {
                // add new variables and constraints

                // first add new constraints (those in increaseD)
                // also set the coefficient for the objective function (third entry)
                for (vi in increaseD) {
                    model.addVar(0.0, 1.0, dValues[vi].toDouble(), GRB.BINARY, "x_${vi}_${dValues[vi]}")
                }
                // add these new variables to the model
                model.update()
                // update the constraints (add the new variables to the existing constraints)
                for (vi in increaseD) {
                    val c = model.getConstrByName("sum vertex$vi")
                    val v = model.getVarByName("x_${vi}_${dValues[vi]}")
                    model.chgCoeff(c, v, 1.0)
                }
                // add the new distance constraints by continuing the BFS for the vertices in increaseD
                for (vi in increaseD) {
                    val term = GRBLinExpr()
                    // continue BFS
                    for (i in 0 until vertQueues[vi].size) {
                        for (ni in neighborhoods[vertQueues[vi].pollFirst()]) {
                            if (ni !in prevVerts[vi]) {
                                // ni has distance d(vi) - 1 to vi
                                prevVerts[vi].add(ni)
                                vertQueues[vi].addLast(ni)
                                if (ni in initialSufficientIndices) {
                                    term.addTerm(1.0, model.getVarByName("x_${ni}_0"))
                                }
                            }
                        }
                    }
                    if (vertQueues[vi].isEmpty()) {
                        finishedBFS[vi] = true
                    }
                    model.addConstr(model.getVarByName("x_${vi}_${dValues[vi] - 1}"), GRB.LESS_EQUAL, term, "ecc $vi ${dValues[vi] - 1}")
                }
                increaseD.clear()
            }
        } while (!solved)

        // retrieve solution
        val bestSet = HashSet<V>(k)
        val bestSetIndices = HashSet<Int>()
        var cent = 0
        for (vi in 0 until vertCount) {
            if (vi in initialSufficientIndices) {
                val v = model.getVarByName("x_${vi}_0")
                if (v.get(GRB.DoubleAttr.X) > 0.99) {
                    bestSet.add(vertexList[vi])
                    bestSetIndices.add(vi)
                }
            }
            for (i in 1 until dValues[vi]) {
                val v = model.getVarByName("x_${vi}_$i")
                if (v.get(GRB.DoubleAttr.X) > 0.99) {
                    cent += i
                }
            }
        }
        model.dispose()
        env.dispose()
        return Triple(bestSet, cent, (vertCount - k).toFloat() / cent.toFloat())
    }
}
