package algos

import gurobi.GRB
import gurobi.GRBEnv
import gurobi.GRBLinExpr
import gurobi.GRBModel
import org.jgrapht.graph.SimpleGraph

class ILPnew<V, E>(graph: SimpleGraph<V, E>) : Algo<V, E>(graph) {

    override fun preparations() {
        createVertexLists()
        calcNeighborhoods()
        createDistanceMatrix()
    }

    override fun calcResult(): Triple<Set<V>, Int, Float> {
        val env = GRBEnv()
        env.set(GRB.IntParam.Threads, 1)
        env.start()

        val model = GRBModel(env)

        // decision variables
        for (i in 0 until vertCount) {
            for (j in 0 until eccentricity[i] + 1) {
                model.addVar(0.0, 1.0, 0.0, GRB.BINARY, "x_" + i.toString() + "_" + j.toString())
            }
            model.addVar(0.0, 1.0, 0.0, GRB.BINARY, "y_" + i.toString())
        }

        model.update()

        // constraints
        val term0 = GRBLinExpr()
        for (i in 0 until vertCount) {
            term0.addTerm(1.0, model.getVarByName("x_" + i.toString() + "_0"))
            val term = GRBLinExpr()
            for (j in 0 until eccentricity[i] + 1) {
                term.addTerm(1.0, model.getVarByName("x_" + i.toString() + "_" + j.toString()))
            }
            model.addConstr(term, GRB.EQUAL, 1.0, "sum vertex" + i.toString())
        }
        model.addConstr(term0, GRB.EQUAL, k.toDouble(), "choose k vertices")

        for (i in 0 until vertCount) {
            for (di in 1..eccentricity[i]) {
                val term = GRBLinExpr()
                for (j in 0 until vertCount) {
                    if (distanceMatrix[i][j] == di) {
                        term.addTerm(1.0, model.getVarByName("x_" + j.toString() + "_0"))
                    }
                }
                model.addConstr(model.getVarByName("x_" + i.toString() + "_" + di.toString()), GRB.LESS_EQUAL, term, "ecc " + i + " " + di)
            }
        }

        // objective function
        val term = GRBLinExpr()
        for (i in 0 until vertCount) {
            for (j in 0 until eccentricity[i] + 1) {
                term.addTerm(j.toDouble(), model.getVarByName("x_" + i.toString() + "_" + j.toString()))
            }
        }
        model.setObjective(term, GRB.MINIMIZE)

        // solve
        model.optimize()

        // retrieve solution
        val bestSet = HashSet<V>(k)
        val bestSetIndices = HashSet<Int>()
        for (i in 0 until vertCount) {
            val v = model.getVarByName("x_" + i.toString() + "_0")
            if (v.get(GRB.DoubleAttr.X) > 0.99) {
                bestSet.add(vertexList[i])
                bestSetIndices.add(i)
            }
        }
        model.dispose()
        env.dispose()

        var cent = 0
        for (i in vertexList.indices) {
            cent += bestSetIndices.minOf { distanceMatrix[i][it] }
        }
        return Triple(bestSet, cent, (vertCount - k).toFloat() / cent.toFloat())
    }
}
