package algos

import gurobi.GRB
import gurobi.GRBEnv
import gurobi.GRBLinExpr
import gurobi.GRBModel
import org.jgrapht.graph.SimpleGraph

class ILPold<V, E>(graph: SimpleGraph<V, E>) : Algo<V, E>(graph) {

    override fun preparations() {
        createVertexLists()
        calcNeighborhoods()
        createDistanceMatrix()
    }

    override fun calcResult(): Triple<Set<V>, Int, Float> {
        val env = GRBEnv()
        env.set(GRB.IntParam.Threads, 1)
        env.start()

        val model = GRBModel(env)
        // model.set(GRB.DoubleParam.TimeLimit, timeLimit) // not needed; this is done outside in main

        // variables
        for (i in 0 until vertCount) {
            for (j in 0 until vertCount) {
                model.addVar(0.0, 1.0, 0.0, GRB.BINARY, "x_" + i.toString() + "_" + j.toString())
            }
            model.addVar(0.0, 1.0, 0.0, GRB.BINARY, "y_" + i.toString())
        }

        model.update()

        // constraints
        val termy = GRBLinExpr()
        for (i in 0 until vertCount) {
            termy.addTerm(1.0, model.getVarByName("y_" + i.toString()))
            val term = GRBLinExpr()
            for (j in 0 until vertCount) {
                term.addTerm(1.0, model.getVarByName("x_" + i.toString() + "_" + j.toString()))
            }
            model.addConstr(term, GRB.EQUAL, 1.0, "sum vertex" + i.toString())
        }
        model.addConstr(termy, GRB.EQUAL, k.toDouble(), "sum vertex y")

        for (i in 0 until vertCount) {
            for (j in 0 until vertCount) {
                model.addConstr(model.getVarByName("x_" + i.toString() + "_" + j.toString()), GRB.LESS_EQUAL, model.getVarByName("y_" + j.toString()), "chossen" + i.toString())
            }
        }

        // objective function
        val term = GRBLinExpr()
        for (i in 0 until vertCount) {
            for (j in 0 until vertCount) {
                term.addTerm(distanceMatrix[i][j].toDouble(), model.getVarByName("x_" + i.toString() + "_" + j.toString()))
            }
        }
        model.setObjective(term, GRB.MINIMIZE)

        // solve
        model.optimize()

        // retrieve solution
        val bestSet = HashSet<V>(k)
        val bestSetIndices = HashSet<Int>()
        for (i in 0 until vertCount) {
            val v = model.getVarByName("y_" + i.toString())
            if (v.get(GRB.DoubleAttr.X) > 0.99) {
                bestSet.add(vertexList[i])
                bestSetIndices.add(i)
            }
        }

        model.dispose()
        env.dispose()

        var cent = 0
        for (i in vertexList.indices) {
            cent += bestSetIndices.minOf { distanceMatrix[i][it] }
        }
        return Triple(bestSet, cent, (vertCount - k).toFloat() / cent.toFloat())
    }
}
